/* -*- c++ -*- */
/*
 * Copyright 2014 <+YOU OR YOUR COMPANY+>.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include <gnuradio/math.h>
#include <pmt/pmt.h>
#include "switch_const_impl.h"

namespace gr {
  namespace mymod {

    switch_const::sptr
    switch_const::make(digital::constellation_sptr constell0,
                       digital::constellation_sptr constell1)
    {
      std::vector<digital::constellation_sptr> constell;
      constell.push_back(constell0);
      constell.push_back(constell1);
      return gnuradio::get_initial_sptr
        (new switch_const_impl(constell));
    }

    switch_const_impl::switch_const_impl(const std::vector<digital::constellation_sptr> &constell)
      : gr::block("switch_const",
                  gr::io_signature::make(1, 1, sizeof(gr_complex)),
                  gr::io_signature::make(0, 0, 0))
    {
      d_constel = constell;
      d_count = 0;
      d_index = 0;
      d_switch_ok = false;

      message_port_register_out(pmt::intern("msg"));
      message_port_register_in(pmt::intern("trig"));
      set_msg_handler(pmt::intern("trig"),
                      boost::bind(&switch_const_impl::trigger, this, _1));

      set_history(2);
    }

    switch_const_impl::~switch_const_impl()
    {
    }

    void
    switch_const_impl::trigger(pmt::pmt_t msg)
    {
      d_switch_ok = true;
    }

    int
    switch_const_impl::general_work(int noutput_items,
                                    gr_vector_int &ninput_items,
                                    gr_vector_const_void_star &input_items,
                                    gr_vector_void_star &output_items)
    {
      const gr_complex *in = (const gr_complex *)input_items[0];
      gr_complex *out = (gr_complex *)output_items[0];

      for(int i = 0; i < noutput_items; i++) {
        double ph0 = gr::fast_atan2f(in[i]);
        double ph1 = gr::fast_atan2f(in[i+1]);
        d_acc_ph(ph0);
        d_acc_fq(ph1 - ph0);
      }

      if(d_switch_ok) {
        if(d_index == 0) { // BPSK
          if(ba::kurtosis(d_acc_fq) < 12) {
            std::cerr << "Found freq. kurtosis " << ba::kurtosis(d_acc_fq)
                      << " < 12" << std::endl;
            std::cerr << "  --> Switching to QPSK" << std::endl;
            d_index = 1;
            d_switch_ok = false;

            d_acc_ph = ba::accumulator_set< double, ba::stats<ba::tag::mean,
                                                              ba::tag::variance,
                                                              ba::tag::kurtosis,
                                                              ba::tag::skewness> >();

            d_acc_fq = ba::accumulator_set< double, ba::stats<ba::tag::mean,
                                                              ba::tag::variance,
                                                              ba::tag::kurtosis,
                                                              ba::tag::skewness> >();

            message_port_pub(pmt::intern("msg"),
                             d_constel[d_index]->as_pmt());
          }
        }
        else { // QPSK
          if(ba::skewness(d_acc_fq) < 1.5) {
            std::cerr << "Found freq. skewness " << ba::skewness(d_acc_fq)
                      << " < 1.5" << std::endl;
            std::cerr << "  --> Switching to BPSK" << std::endl;
            d_index = 0;
            d_switch_ok = false;

            d_acc_ph = ba::accumulator_set< double, ba::stats<ba::tag::mean,
                                                              ba::tag::variance,
                                                              ba::tag::kurtosis,
                                                              ba::tag::skewness> >();

            d_acc_fq = ba::accumulator_set< double, ba::stats<ba::tag::mean,
                                                              ba::tag::variance,
                                                              ba::tag::kurtosis,
                                                              ba::tag::skewness> >();

            message_port_pub(pmt::intern("msg"),
                             d_constel[d_index]->as_pmt());
          }
        }
      }

//      d_count++;
//      if(d_count % 4000 == 0) {
//        std::cerr << "Phase: mean: " << ba::mean(d_acc_ph)
//                  << "   kurt: " << ba::kurtosis(d_acc_ph)
//                  << "   skew: " << ba::skewness(d_acc_ph)
//                  << "   std:  " << sqrt(ba::variance(d_acc_ph)) << std::endl;
//
//        std::cerr << "Freq: mean: " << ba::mean(d_acc_fq)
//                  << "   kurt: " << ba::kurtosis(d_acc_fq)
//                  << "   skew: " << ba::skewness(d_acc_fq)
//                  << "   std:  " << sqrt(ba::variance(d_acc_fq)) << std::endl;
//      }

      consume_each(noutput_items);
      return noutput_items;
    }

  } /* namespace mymod */
} /* namespace gr */
